from __future__ import annotations

import os
import platform
import re
import shutil
import subprocess
import typing as t
from pathlib import Path
from time import perf_counter_ns

import click
import mock
import pytest
import yaml
from structlog.stdlib import get_logger

import meltano
from asserts import assert_cli_runner
from fixtures.utils import cd
from meltano.cli import cli, handle_meltano_error
from meltano.cli.utils import CliError
from meltano.core.error import EmptyMeltanoFileException, MeltanoError
from meltano.core.logging.utils import setup_logging
from meltano.core.project import PROJECT_ENVIRONMENT_ENV, PROJECT_READONLY_ENV, Project

if t.TYPE_CHECKING:
    from fixtures.cli import MeltanoCliRunner

ANSI_RE = re.compile(r"\033\[[;?0-9]*[a-zA-Z]")


class TestCli:
    @pytest.fixture()
    def test_cli_project(self, tmp_path: Path, project_init_service):
        """Return the non-activated project."""
        os.chdir(tmp_path)
        project = project_init_service.init(activate=False)
        Project._default = None
        try:
            yield project
        finally:
            Project.deactivate()
            shutil.rmtree(project.root)

    @pytest.fixture()
    def deactivate_project(self) -> None:
        Project.deactivate()

    @pytest.fixture()
    def empty_project(
        self,
        empty_meltano_yml_dir,
        pushd,  # noqa: ARG002
    ):
        project = Project(empty_meltano_yml_dir)
        try:
            yield project
        finally:
            Project.deactivate()

    @pytest.mark.order(0)
    def test_activate_project(self, test_cli_project, cli_runner, pushd) -> None:
        project = test_cli_project

        pushd(project.root)
        cli_runner.invoke(cli, ["hub", "ping"])

        assert Project._default is not None
        assert Project._default.root == project.root
        assert Project._default.readonly is False

    @pytest.mark.order(1)
    def test_empty_meltano_yml_project(self, empty_project, cli_runner, pushd) -> None:
        pushd(empty_project.root)
        with pytest.raises(EmptyMeltanoFileException):
            cli_runner.invoke(cli, ["config"], catch_exceptions=False)

    @pytest.mark.order(2)
    def test_activate_project_readonly_env(
        self,
        test_cli_project,
        cli_runner,
        pushd,
        monkeypatch,
    ) -> None:
        monkeypatch.setenv(PROJECT_READONLY_ENV, "true")
        assert Project._default is None
        pushd(test_cli_project.root)
        cli_runner.invoke(cli, ["hub", "ping"])
        assert Project._default.readonly

    @pytest.mark.order(2)
    def test_activate_project_readonly_dotenv(
        self,
        test_cli_project,
        cli_runner,
        pushd,
    ) -> None:
        test_cli_project.settings.set("project_readonly", value=True)
        assert Project._default is None
        pushd(test_cli_project.root)
        cli_runner.invoke(cli, ["hub", "ping"])
        assert Project._default.readonly

    def test_environment_precedence(
        self,
        project: Project,
        pushd,
        cli_runner: MeltanoCliRunner,
        monkeypatch: pytest.MonkeyPatch,
    ) -> None:
        pushd(project.root)
        monkeypatch.delenv(PROJECT_ENVIRONMENT_ENV, raising=False)
        environment_names = {
            name: f"env_set_from_{name}" for name in ("dotenv", "cli_option", "env_var")
        }
        with mock.patch(
            "meltano.core.project.Project.dotenv_env",
            new_callable=mock.PropertyMock,
            return_value={PROJECT_ENVIRONMENT_ENV: environment_names["dotenv"]},
        ):
            args = ("invoke", "tap-mock")
            results = {
                "dotenv": cli_runner.invoke(cli, args),
                "cli_option": cli_runner.invoke(
                    cli,
                    (f"--environment={environment_names['cli_option']}", *args),
                ),
                "env_var": cli_runner.invoke(
                    cli,
                    args,
                    env={PROJECT_ENVIRONMENT_ENV: environment_names["env_var"]},
                ),
            }
        for source, name in environment_names.items():
            assert results[source].exit_code
            assert (
                results[source].exception.args[0]
                == f"Environment {name!r} was not found."
            )

    def test_version(self, cli_runner) -> None:
        cli_version = cli_runner.invoke(cli, ["--version"])

        assert cli_version.output == f"meltano, version {meltano.__version__}\n"

    @pytest.mark.usefixtures("deactivate_project")
    def test_default_environment_is_activated(
        self,
        project_files_cli,
        cli_runner,
        pushd,
    ) -> None:
        pushd(project_files_cli.root)
        cli_runner.invoke(
            cli,
            ["test"],
        )
        assert Project._default.environment.name == "test-meltano-environment"

    @pytest.mark.usefixtures("deactivate_project")
    def test_environment_flag_overrides_default(
        self,
        project_files_cli,
        cli_runner,
        pushd,
    ) -> None:
        pushd(project_files_cli.root)
        cli_runner.invoke(
            cli,
            ["--environment", "test-subconfig-2-yml", "test"],
        )

        assert Project._default.environment.name == "test-subconfig-2-yml"

    @pytest.mark.usefixtures("deactivate_project")
    def test_environment_variable_overrides_default(
        self,
        project_files_cli,
        cli_runner,
        pushd,
        monkeypatch,
    ) -> None:
        monkeypatch.setenv("MELTANO_ENVIRONMENT", "test-subconfig-2-yml")
        pushd(project_files_cli.root)
        cli_runner.invoke(
            cli,
            ["test"],
        )
        assert Project._default.environment.name == "test-subconfig-2-yml"

    @pytest.mark.usefixtures("deactivate_project")
    def test_lower_null_environment_overrides_default(
        self,
        project_files_cli,
        cli_runner,
        pushd,
    ) -> None:
        pushd(project_files_cli.root)
        cli_runner.invoke(
            cli,
            ["--environment", "null", "hub", "ping"],
        )
        assert Project._default.environment is None

    @pytest.mark.usefixtures("deactivate_project")
    def test_upper_null_environment_overrides_default(
        self,
        project_files_cli,
        cli_runner,
        pushd,
    ) -> None:
        pushd(project_files_cli.root)
        cli_runner.invoke(
            cli,
            ["--environment", "NULL", "hub", "ping"],
        )
        assert Project._default.environment is None

    @pytest.mark.usefixtures("deactivate_project")
    def test_no_environment_overrides_default(
        self,
        project_files_cli,
        cli_runner,
        pushd,
    ) -> None:
        pushd(project_files_cli.root)
        cli_runner.invoke(
            cli,
            ["--no-environment", "hub", "ping"],
        )
        assert Project._default.environment is None

    @pytest.mark.usefixtures("deactivate_project")
    def test_no_environment_and_null_environment_overrides_default(
        self,
        project_files_cli,
        cli_runner,
        pushd,
    ) -> None:
        pushd(project_files_cli.root)
        cli_runner.invoke(
            cli,
            ["--no-environment", "--environment", "null", "hub", "ping"],
        )
        assert Project._default.environment is None

    def test_handle_meltano_error(self) -> None:
        exception = MeltanoError(reason="This failed", instruction="Try again")
        with pytest.raises(CliError, match="This failed. Try again."):
            handle_meltano_error(exception)

    @pytest.mark.usefixtures("pushd")
    def test_cwd_option(
        self,
        cli_runner,
        test_cli_project,
        tmp_path: Path,
    ) -> t.NoReturn:
        project = test_cli_project
        with cd(project.root_dir()):
            assert_cli_runner(cli_runner.invoke(cli, ("dragon",)))
            assert Path().resolve() == project.root_dir()

        with cd(project.root_dir()):
            assert_cli_runner(
                cli_runner.invoke(cli, ("--cwd", str(tmp_path), "dragon")),
            )
            assert Path().resolve() == tmp_path

        with cd(project.root_dir()):
            filepath = tmp_path / "file.txt"
            filepath.touch()
            with pytest.raises(click.BadParameter, match="is a file"):
                raise cli_runner.invoke(
                    cli,
                    ("--cwd", str(filepath), "dragon"),
                ).exception.__context__

        with cd(project.root_dir()):
            dirpath = tmp_path / "subdir"
            with pytest.raises(click.BadParameter, match="does not exist"):
                raise cli_runner.invoke(
                    cli,
                    ("--cwd", str(dirpath), "dragon"),
                ).exception.__context__

        with cd(project.root_dir()):
            dirpath.mkdir()
            assert_cli_runner(cli_runner.invoke(cli, ("--cwd", str(dirpath), "dragon")))
            assert Path().resolve() == dirpath

    @pytest.mark.parametrize(
        "command_args",
        (
            ("invoke", "example"),
            ("config", "example"),
            ("job", "list"),
            ("environment", "list"),
            ("add", "utility", "example"),
        ),
    )
    def test_error_msg_outside_project(
        self,
        tmp_path: Path,
        command_args: tuple[str, ...],
    ) -> None:
        # Unless this test runs before every test that uses a project, we
        # cannot use `cli_runner` to test this because the code path taken
        # differs after any project has been found.

        # I tried working around this by switching to an empty directory,
        # calling `Project.deactivate`, using `mock.patch` on various relevant
        # functions, and more, but nothing I did resulted in the proper code
        # path being taken. Also it seemed like a fragile approach.

        # Using a subprocess should be robust, but requires the version of
        # Meltano you want to test be the one that is installed in the active
        # Python environment. This is not the only test that requires this.
        assert (
            "must be run inside a Meltano project"
            in subprocess.run(
                ("meltano", *command_args),
                text=True,
                stderr=subprocess.PIPE,
                cwd=tmp_path,
            ).stderr
        )


def _get_dummy_logging_config(*, colors=True):
    return {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "default": {
                "()": "meltano.core.logging.console_log_formatter",
                "colors": colors,
            },
        },
        "handlers": {
            "console": {
                "class": "logging.StreamHandler",
                "formatter": "default",
            },
        },
        "loggers": {
            "meltano.cli.dummy": {
                "handlers": ["console"],
                "level": "DEBUG",
            },
        },
    }


@pytest.mark.skipif(
    platform.system() == "Windows",
    reason="Windows terminal support for ANSI escape sequences is limited.",
)
class TestCliColors:
    TEST_TEXT = "This is a test"

    @pytest.mark.parametrize(
        ("env", "log_config", "cli_colors_expected", "log_colors_expected"),
        (
            pytest.param(
                {},
                None,
                True,
                True,
                id="colors-enabled-by-default",
            ),
            pytest.param(
                {
                    "NO_COLOR": "1",
                },
                _get_dummy_logging_config(colors=True),
                False,
                False,
                id="all-colors-disabled--no-color-precedence",
            ),
            pytest.param(
                {},
                _get_dummy_logging_config(colors=False),
                True,
                False,
                id="cli-colors-enabled-log-colors-disabled",
            ),
            pytest.param(
                {
                    "NO_COLOR": "1",
                },
                None,
                False,
                False,
                id="colors-disabled-by-1-no-color-env",
            ),
            pytest.param(
                {
                    "NO_COLOR": "TRUE",
                },
                None,
                False,
                False,
                id="colors-disabled-by-TRUE-no-color-env",
            ),
            pytest.param(
                {
                    "NO_COLOR": "t",
                },
                None,
                False,
                False,
                id="colors-disabled-by-t-no-color-env",
            ),
            pytest.param(
                {
                    "NO_COLOR": "f",
                },
                None,
                True,
                True,
                id="colors-not-disabled-by-f-no-color-env",
            ),
            pytest.param(
                {
                    "NO_COLOR": "FALSE",
                },
                None,
                True,
                True,
                id="colors-not-disabled-by-FALSE-no-color-env",
            ),
            pytest.param(
                {
                    "NO_COLOR": "NOT_A_BOOLEAN",
                },
                None,
                True,
                True,
                id="colors-not-disabled-by-invalid-no-color-env",
            ),
        ),
    )
    def test_no_color(
        self,
        cli_runner,
        env,
        log_config,
        cli_colors_expected,
        log_colors_expected,
        tmp_path,
        monkeypatch,
    ) -> None:
        monkeypatch.delenv("NO_COLOR", raising=False)
        styled_text = click.style(self.TEST_TEXT, fg="red")

        if log_config:
            log_config_path = tmp_path / "logging.yml"
            log_config_path.write_text(yaml.dump(log_config))
        else:
            log_config_path = None

        @click.command("dummy")
        def dummy_command() -> None:
            setup_logging(None, "DEBUG", log_config_path)
            logger = get_logger("meltano.cli.dummy")
            logger.info(self.TEST_TEXT)
            click.echo(styled_text)

        cli.add_command(dummy_command)

        expected_text = styled_text if cli_colors_expected else self.TEST_TEXT

        with cli_runner.isolated_filesystem():
            result = cli_runner.invoke(cli, ["dummy"], color=True, env=env)
            assert result.exit_code == 0, result.exception
            assert result.output.strip() == expected_text
            assert bool(ANSI_RE.match(result.stderr)) is log_colors_expected
            assert result.exception is None


class TestLargeConfigProject:
    @pytest.mark.usefixtures("large_config_project")
    def test_list_config_performance(self, cli_runner) -> None:
        start = perf_counter_ns()
        assert (
            cli_runner.invoke(
                cli,
                ["--no-environment", "config", "target-with-large-config", "list"],
            ).exit_code
            == 0
        )
        duration_ns = perf_counter_ns() - start
        # Ensure the large config can be processed in less than 25 seconds
        assert duration_ns < 25000000000



from __future__ import annotations

import typing as t
from unittest.mock import AsyncMock, patch

import pytest
import yaml

from meltano.core.plugin_install_service import (
    PluginInstallReason,
    PluginInstallService,
    get_pip_install_args,
)

if t.TYPE_CHECKING:
    from meltano.core.project import Project


class TestPluginInstallService:
    @pytest.fixture(
        params=({}, {"parallelism": -1}, {"parallelism": 2}),
        ids=("default", "-p=-1", "-p=2"),
    )
    def subject(self, project: Project, request):
        with project.meltanofile.open("w") as file:
            file.write(
                yaml.dump(
                    {
                        "plugins": {
                            "extractors": [
                                {
                                    "name": "tap-gitlab",
                                    "namespace": "tap_gitlab",
                                    "pip_url": "git+https://gitlab.com/meltano/tap-gitlab.git",
                                },
                                {
                                    "name": "tap-gitlab--child-1",
                                    "inherit_from": "tap-gitlab",
                                },
                            ],
                            "loaders": [
                                {
                                    "name": "target-csv",
                                    "namespace": "target_csv",
                                    "pip_url": "git+https://gitlab.com/meltano/target-csv.git",
                                },
                            ],
                        },
                    },
                ),
            )
        project.refresh()
        return PluginInstallService(project, **request.param)

    def test_default_init_should_not_fail(self, subject) -> None:
        assert subject

    def test_remove_duplicates(self, subject) -> None:
        states, deduped_plugins = subject.remove_duplicates(
            plugins=subject.project.plugins.plugins(),
            reason=PluginInstallReason.INSTALL,
        )

        assert len(states) == 1
        assert states[0].plugin.name == "tap-gitlab--child-1"
        assert states[0].skipped

        assert len(deduped_plugins) == 2
        assert [plugin.name for plugin in deduped_plugins] == [
            "tap-gitlab",
            "target-csv",
        ]

    @pytest.mark.slow()
    async def test_install_all(self, subject) -> None:
        all_plugins = await subject.install_all_plugins()
        assert len(all_plugins) == 3

        assert all_plugins[2].plugin.name == "target-csv"
        assert all_plugins[2].successful

        # test inherited plugins behaviors
        assert all_plugins[0].plugin.name == "tap-gitlab--child-1"
        assert all_plugins[0].successful
        assert all_plugins[0].skipped

        assert all_plugins[1].plugin.name == "tap-gitlab"
        assert all_plugins[1].successful
        assert not all_plugins[1].skipped

        assert all_plugins[0].plugin.venv_name == all_plugins[1].plugin.venv_name
        assert all_plugins[0].plugin.executable == all_plugins[1].plugin.executable

    def test_get_quoted_pip_install_args(self, project) -> None:
        with project.meltanofile.open("w") as file:
            file.write(
                yaml.dump(
                    {
                        "plugins": {
                            "extractors": [
                                {
                                    "name": "tap-gitlab",
                                    "namespace": "tap_gitlab",
                                    "pip_url": "'tap-gitlab @ git+https://gitlab.com/meltano/tap-gitlab.git' python-json-logger",  # noqa: E501
                                },
                            ],
                        },
                    },
                ),
            )
        project.refresh()
        plugin = next(project.plugins.plugins())
        assert get_pip_install_args(project, plugin) == [
            "tap-gitlab @ git+https://gitlab.com/meltano/tap-gitlab.git",
            "python-json-logger",
        ]

    @patch("meltano.core.venv_service.VenvService.install_pip_args", AsyncMock())
    @pytest.mark.usefixtures("reset_project_context")
    async def test_auto_install(
        self,
        project: Project,
        subject: PluginInstallService,
    ) -> None:
        plugin = next(project.plugins.plugins())
        state = await subject.install_plugin_async(
            plugin,
            reason=PluginInstallReason.AUTO,
        )

        assert not state.skipped, "Expected plugin with no venv to be installed"

        state = await subject.install_plugin_async(
            plugin,
            reason=PluginInstallReason.AUTO,
        )

        assert (
            state.skipped
        ), "Expected plugin with venv and matching fingerprint to not be installed"

        plugin.pip_url = "changed"
        state = await subject.install_plugin_async(
            plugin,
            reason=PluginInstallReason.AUTO,
        )

        assert (
            not state.skipped
        ), "Expected plugin with venv and non-matching fingerprint to be installed"

        plugin.pip_url = "$MISSING_ENV_VAR"
        state = await subject.install_plugin_async(
            plugin,
            reason=PluginInstallReason.AUTO,
        )

        assert (
            state.skipped
        ), "Expected plugin with missing env var in pip URL to not be installed"



"""Snowplow contexts for Meltano telemetry."""

from __future__ import annotations

from meltano.core.tracking.contexts.cli import CliContext, CliEvent
from meltano.core.tracking.contexts.environment import (
    EnvironmentContext,
    environment_context,
)
from meltano.core.tracking.contexts.exception import ExceptionContext
from meltano.core.tracking.contexts.plugins import PluginsTrackingContext
from meltano.core.tracking.contexts.project import ProjectContext



from __future__ import annotations

import json
import platform
import typing as t

import mock
import pytest

from asserts import assert_cli_runner
from meltano.cli import cli, state
from meltano.core.utils import merge

if t.TYPE_CHECKING:
    from pathlib import Path

unconventional_state_ids = [
    "unconventional",
    "dev:tap-and-target",
    "tap-mock-to-target-mock",
    "staging:",
    "staging:tap-mock-to-",
    "dev:-to-target-mock",
    "dev:tap-to-target:",
    "dev:tap-to-target:suffix:",
]

conventional_state_ids = [
    "dev:tap-mock-to-target-mock",
    "dev:tap-mock-to-target-mock:suffix",
    "staging:mock-to-mock",
]


class TestCliState:
    @pytest.mark.parametrize("state_id", unconventional_state_ids)
    def test_state_service_from_state_id_returns_none_non_convention(
        self,
        project,
        state_id,
    ) -> None:
        assert state.state_service_from_state_id(project, state_id) is None

    @pytest.mark.parametrize("state_id", conventional_state_ids)
    def test_state_service_from_state_id_returns_state_service_convention(
        self,
        project,
        state_id,
    ) -> None:
        with mock.patch(
            "meltano.cli.state.BlockParser",
            autospec=True,
        ) as mock_block_parser:
            state.state_service_from_state_id(project, state_id)
            args = state_id.split(":")[1].split("-to-")
            assert args in mock_block_parser.call_args.args

    @staticmethod
    def get_result_set(result):
        result_set = set(result.stdout.split("\n"))
        result_set.remove("")
        return result_set

    @pytest.mark.usefixtures("project")
    def test_list(self, state_ids, state_service, cli_runner) -> None:
        with mock.patch("meltano.cli.state.StateService", return_value=state_service):
            result = cli_runner.invoke(cli, ["state", "list"])
        assert_cli_runner(result)
        assert self.get_result_set(result) == set(state_ids)

    @pytest.fixture()
    def patterns_with_expected_results(self, state_ids):
        return [
            (
                "test:*",
                set(
                    filter(
                        lambda state_id: state_id.startswith("test:"),
                        list(state_ids),
                    ),
                ),
            ),
            ("*-to-*", set(state_ids)),
            ("multiple-complete", set()),
            (
                "*multiple-complete",
                set(
                    filter(
                        lambda state_id: state_id.endswith("multiple-complete"),
                        list(state_ids),
                    ),
                ),
            ),
        ]

    def test_list_pattern(
        self,
        state_service,
        cli_runner,
        patterns_with_expected_results,
    ) -> None:
        with mock.patch("meltano.cli.state.StateService", return_value=state_service):
            for pattern, expected_result in patterns_with_expected_results:
                result = cli_runner.invoke(cli, ["state", "list", "--pattern", pattern])
                assert_cli_runner(result)
                assert self.get_result_set(result) == expected_result

    def test_set_from_string(
        self,
        state_service,
        state_ids,
        payloads,
        cli_runner,
    ) -> None:
        with mock.patch("meltano.cli.state.StateService", return_value=state_service):
            for state_id in state_ids:
                for state_payload in payloads.mock_state_payloads:
                    result = cli_runner.invoke(
                        cli,
                        [
                            "state",
                            "set",
                            "--force",
                            state_id,
                            json.dumps(state_payload),
                        ],
                    )
                    assert_cli_runner(result)
                    assert state_service.get_state(state_id) == state_payload

    def test_set_from_file(
        self,
        tmp_path: Path,
        state_service,
        state_ids,
        payloads,
        cli_runner,
    ) -> None:
        with mock.patch("meltano.cli.state.StateService", return_value=state_service):
            for idx_i, state_id in enumerate(state_ids):
                for idx_j, state_payload in enumerate(payloads.mock_state_payloads):
                    filepath = tmp_path / f"state-file-{idx_i}-{idx_j}.json"
                    with filepath.open("w+") as state_file:
                        json.dump(state_payload, state_file)
                    result = cli_runner.invoke(
                        cli,
                        ["state", "set", "--force", state_id, "--input-file", filepath],
                    )
                    assert_cli_runner(result)
                    assert state_service.get_state(state_id) == state_payload

    def test_merge_from_string(self, state_service, state_ids, cli_runner) -> None:
        with mock.patch("meltano.cli.state.StateService", return_value=state_service):
            job_pairs = [
                (state_ids[idx], state_ids[idx + 1])
                for idx in range(0, len(state_ids) - 1, 2)
            ]
            for job_src, job_dst in job_pairs:
                job_src_state = state_service.get_state(job_src)
                job_dst_state = state_service.get_state(job_dst)
                result = cli_runner.invoke(
                    cli,
                    ["state", "merge", job_dst, json.dumps(job_src_state)],
                )
                assert_cli_runner(result)
                assert state_service.get_state(job_dst) == merge(
                    job_src_state,
                    job_dst_state,
                )

    @pytest.mark.usefixtures("payloads")
    def test_merge_from_file(
        self,
        tmp_path: Path,
        state_service,
        state_ids,
        cli_runner,
    ) -> None:
        if platform.system() == "Windows":
            pytest.xfail(
                "Fails on Windows: https://github.com/meltano/meltano/issues/3444",
            )
        with mock.patch("meltano.cli.state.StateService", return_value=state_service):
            job_pairs = [
                (state_ids[idx], state_ids[idx + 1])
                for idx in range(0, len(state_ids) - 1, 2)
            ]
            for job_src, job_dst in job_pairs:
                job_src_state = state_service.get_state(job_src)
                job_dst_state = state_service.get_state(job_dst)
                filepath = tmp_path / f"{job_src}-{job_dst}"
                with filepath.open("w+") as state_file:
                    json.dump(job_src_state, state_file)
                result = cli_runner.invoke(
                    cli,
                    ["state", "merge", job_dst, "--input-file", filepath],
                )
                assert_cli_runner(result)
                assert state_service.get_state(job_dst) == merge(
                    job_src_state,
                    job_dst_state,
                )

    def test_merge_from_job(self, state_service, state_ids, cli_runner) -> None:
        with mock.patch("meltano.cli.state.StateService", return_value=state_service):
            job_pairs = [
                (state_ids[idx], state_ids[idx + 1])
                for idx in range(0, len(state_ids) - 1, 2)
            ]
            for job_src, job_dst in job_pairs:
                job_state_src = state_service.get_state(job_src)
                job_state_dst = state_service.get_state(job_dst)
                merged_state = merge(job_state_src, job_state_dst)
                result = cli_runner.invoke(
                    cli,
                    ["state", "merge", "--from-state-id", job_src, job_dst],
                )
                assert_cli_runner(result)
                assert state_service.get_state(job_dst) == merged_state

    def test_copy_over_existing(self, state_service, state_ids, cli_runner) -> None:
        with mock.patch("meltano.cli.state.StateService", return_value=state_service):
            job_pairs = [
                (state_ids[idx], state_ids[idx + 1])
                for idx in range(0, len(state_ids) - 1, 2)
            ]
            for job_src, job_dst in job_pairs:
                job_src_state = state_service.get_state(job_src)
                result = cli_runner.invoke(
                    cli,
                    ["state", "copy", job_src, job_dst, "--force"],
                )
                assert_cli_runner(result)
                assert state_service.get_state(job_dst) == job_src_state

    def test_copy_to_new(self, state_service, state_ids, cli_runner) -> None:
        with mock.patch("meltano.cli.state.StateService", return_value=state_service):
            for job_src_id in state_ids:
                job_src_state = state_service.get_state(job_src_id)
                job_dst_id = f"{job_src_id}-test-copy"
                result = cli_runner.invoke(
                    cli,
                    ["state", "copy", job_src_id, job_dst_id, "--force"],
                )
                assert_cli_runner(result)
                assert state_service.get_state(job_dst_id) == job_src_state

    def test_move(self, state_service, state_ids, cli_runner) -> None:
        with mock.patch("meltano.cli.state.StateService", return_value=state_service):
            job_pairs = [
                (state_ids[idx], state_ids[idx + 1])
                for idx in range(0, len(state_ids) - 1, 2)
            ]
            for job_src, job_dst in job_pairs:
                job_src_state = state_service.get_state(job_src)
                result = cli_runner.invoke(
                    cli,
                    ["state", "move", job_src, job_dst, "--force"],
                )
                assert_cli_runner(result)
                assert not state_service.get_state(job_src)
                assert state_service.get_state(job_dst) == job_src_state

    def test_get(
        self,
        state_service,
        cli_runner,
        state_ids_with_expected_states,
    ) -> None:
        with mock.patch("meltano.cli.state.StateService", return_value=state_service):
            for state_id, expected_state in state_ids_with_expected_states:
                result = cli_runner.invoke(cli, ["state", "get", state_id])
                assert_cli_runner(result)
                assert json.loads(result.stdout) == expected_state

    def test_clear(self, state_service, cli_runner, state_ids) -> None:
        with mock.patch("meltano.cli.state.StateService", return_value=state_service):
            for state_id in state_ids:
                result = cli_runner.invoke(cli, ["state", "clear", "--force", state_id])
                assert_cli_runner(result)
                job_state = state_service.get_state(state_id)
                assert (not job_state) or (not job_state.get("singer_state"))

    def test_clear_prompt(self, state_service, cli_runner, state_ids) -> None:
        with mock.patch("meltano.cli.state.StateService", return_value=state_service):
            for state_id in state_ids:
                result = cli_runner.invoke(cli, ["state", "clear", state_id], input="n")
                assert result.exit_code == 1

                job_state = state_service.get_state(state_id)
                assert "singer_state" in job_state

                result = cli_runner.invoke(cli, ["state", "clear", state_id], input="y")
                assert_cli_runner(result)

                job_state = state_service.get_state(state_id)
                assert (not job_state) or (not job_state.get("singer_state"))



"""Store and retrieve Meltano manifests."""

# TODO: This module should handle reading/writing manifest files. Currently
# they're written by `meltano.cli.compile`, and not read by anything. Writing
# them is fairly simple, but reading is tricker because we need to decide
# whether we should read a manifest file from disk (if it exists), or generate
# a new one.



"""A low-level block type for use with `BlockSet` implementations."""

from __future__ import annotations

import typing as t
from abc import ABCMeta, abstractmethod

if t.TYPE_CHECKING:
    from asyncio import StreamWriter, Task

    from meltano.core.logging.utils import SubprocessOutputWriter


class IOBlock(metaclass=ABCMeta):
    """A block that consumes, produces, or both (i.e. transforms) output.

    Underlying implementation could be subprocesses (e.g. Singer Plugins), or
    stream transformers, or any class that satisfies the `IOBlock` interface.
    """

    @property
    @abstractmethod
    def stdin(self) -> StreamWriter | None:
        """Get the `StreamWriter` that should be used for writes, if any.

        Raises:
            NotImplementedError

        Returns:
            StreamWriter
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def consumer(self) -> bool:
        """Indicate whether this block is a consumer and requires input."""
        raise NotImplementedError

    @property
    @abstractmethod
    def producer(self) -> bool:
        """Indicate whether this block is a producer of output."""
        raise NotImplementedError

    @property
    @abstractmethod
    def string_id(self) -> str:
        """Return a string identifier for this block."""
        raise NotImplementedError

    @property
    @abstractmethod
    def has_state(self) -> bool:
        """Indicate whether this block has persistent state between runs."""
        raise NotImplementedError

    @abstractmethod
    def stdout_link(self, dst: SubprocessOutputWriter) -> None:
        """Use stdout_link to instruct block to link/write stdout content to dst.

        Args:
            dst: SubprocessOutputWriter

        Raises:
            NotImplementedError
        """
        raise NotImplementedError

    @abstractmethod
    def stderr_link(self, dst: SubprocessOutputWriter) -> None:
        """Use stderr_link to instruct block to link/write stderr content to dst.

        Args:
            dst: SubprocessOutputWriter

        Raises:
            NotImplementedError
        """
        raise NotImplementedError

    @abstractmethod
    async def start(self) -> None:
        """Start the block.

        Whatever that might entail (spwaning a process, spinning up a async
        task that will handle transforms, etc).

        Raises:
            NotImplementedError
        """
        raise NotImplementedError

    @abstractmethod
    async def stop(self, *, kill: bool = True) -> None:
        """Stop a block.

        Args:
            kill: Whether to send a SIGKILL. If false, a SIGTERM is sent.

        Raises:
            NotImplementedError
        """
        raise NotImplementedError

    @abstractmethod
    def proxy_stdout(self) -> Task:
        """Start proxying stdout to the linked stdout destinations.

        Returns: Future of the proxy task.
        """
        raise NotImplementedError

    @abstractmethod
    def proxy_stderr(self) -> Task:
        """Start proxying stderr to the linked stderr destinations.

        Returns: Future of the proxy task.
        """
        raise NotImplementedError

    @abstractmethod
    def proxy_io(self) -> tuple[Task, Task]:
        """Start proxying stdout AND stderr to the linked destinations.

        Returns:
            proxy_stdout Task and proxy_stderr Task
        """
        stdout = self.proxy_stdout()
        stderr = self.proxy_stderr()
        return stdout, stderr

    @abstractmethod
    async def pre(self, context: object) -> None:
        """Execute pre-start tasks.

        Args:
            context: invocation context to use for this execution.
        """

    @abstractmethod
    async def post(self) -> None:
        """Execute post-stop tasks."""

    @abstractmethod
    async def close_stdin(self) -> None:
        """Close the underlying stdin if the block is a consumer."""



from __future__ import annotations

import typing as t

import pytest
from mock import Mock

from meltano.core.plugin import PluginType
from meltano.core.validation_service import ValidationsRunner

if t.TYPE_CHECKING:
    from meltano.core.project import Project


class MockValidationsRunner(ValidationsRunner):
    async def run_test(self, name: str) -> int:  # noqa: ARG002
        return 1


class TestValidationsRunner:
    @pytest.mark.asyncio()
    async def test_run_all(self, session, dbt, plugin_invoker_factory) -> None:
        invoker = plugin_invoker_factory(dbt)
        runner = MockValidationsRunner(
            invoker,
            {
                "test": True,
                "other-test": False,
                "skipped-test": False,
            },
        )
        assert await runner.run_all(session) == {"test": 1}

        runner.select_test("other-test")
        assert await runner.run_all(session) == {"test": 1, "other-test": 1}

        runner.select_all()
        assert await runner.run_all(session) == {
            "test": 1,
            "other-test": 1,
            "skipped-test": 1,
        }

        noop_runner = MockValidationsRunner(invoker, {})
        noop_runner.invoker = Mock()
        result = await noop_runner.run_all(session)
        assert not result
        assert noop_runner.invoker.call_count == 0

    @pytest.mark.order(after="test_run_all")
    def test_collect_tests(self, project: Project) -> None:
        collected = MockValidationsRunner.collect(project, select_all=False)

        assert collected["dbt"].invoker.plugin.type == PluginType.TRANSFORMERS
        assert "test" in collected["dbt"].tests_selection
        assert not collected["dbt"].tests_selection["test"]

        collected["dbt"].select_all()
        assert collected["dbt"].tests_selection["test"]



from __future__ import annotations  # noqa: D100

import json
import typing as t

from meltano.core.plugin import PluginType
from meltano.core.plugin.error import PluginExecutionError
from meltano.core.plugin.settings_service import PluginSettingsService
from meltano.core.plugin.singer.catalog import ListSelectedExecutor
from meltano.core.plugin_invoker import invoker_factory

if t.TYPE_CHECKING:
    from sqlalchemy.orm.session import Session

    from meltano.core.plugin.base import PluginRef
    from meltano.core.plugin.project_plugin import ProjectPlugin
    from meltano.core.project import Project


class SelectService:  # noqa: D101
    def __init__(self, project: Project, extractor: str):
        """Initialize a `SelectService` instance.

        Args:
            project: The Meltano project being operated on.
            extractor: The name of the extractor plugin.
        """
        self.project = project
        self._extractor = self.project.plugins.find_plugin(
            extractor,
            PluginType.EXTRACTORS,
        )

    @property
    def extractor(self) -> ProjectPlugin:
        """Retrieve extractor ProjectPlugin object."""
        return self._extractor

    @property
    def current_select(self) -> list[str]:
        """Get the current select patterns.

        Returns:
            The current select pattern.
        """
        plugin_settings_service = PluginSettingsService(self.project, self.extractor)
        return plugin_settings_service.get("_select")

    async def load_catalog(self, session: Session, *, refresh: bool = False) -> dict:
        """Load the catalog."""
        invoker = invoker_factory(self.project, self.extractor)

        if refresh:
            invoker.settings_service.config_override["_use_cached_catalog"] = False

        async with invoker.prepared(session):
            catalog_json = await invoker.dump("catalog")

        return json.loads(catalog_json)

    async def list_all(
        self,
        session: Session,
        *,
        refresh: bool = False,
    ) -> ListSelectedExecutor:
        """List all select."""
        try:
            catalog = await self.load_catalog(session, refresh=refresh)
        except FileNotFoundError as err:
            raise PluginExecutionError(
                "Could not find catalog. Verify that the tap supports discovery "  # noqa: EM101
                "mode and advertises the `discover` capability as well as either "
                "`catalog` or `properties`",
            ) from err

        list_all = ListSelectedExecutor()

        # TODO: revisit the visit_with decorator when mypy has better support
        # for class decorators
        # https://github.com/python/mypy/issues/3135
        list_all.visit(catalog)  # type: ignore[attr-defined]

        return list_all

    def update(
        self,
        entities_filter: str,
        attributes_filter: str,
        exclude: bool,  # noqa: FBT001
        *,
        remove: bool = False,
    ) -> None:
        """Update plugins' select patterns."""
        plugin: PluginRef

        if self.project.environment is None:
            plugin = self.extractor
        else:
            plugin = self.project.environment.get_plugin_config(
                self.extractor.type,
                self.extractor.name,
            )

        this_pattern = self._get_pattern_string(
            entities_filter,
            attributes_filter,
            exclude,
        )
        patterns = plugin.extras.get("select", [])
        if remove:
            patterns.remove(this_pattern)
        else:
            patterns.append(this_pattern)
        plugin.extras["select"] = patterns

        if self.project.environment is None:
            self.project.plugins.update_plugin(plugin)
        else:
            self.project.plugins.update_environment_plugin(plugin)

    @staticmethod
    def _get_pattern_string(
        entities_filter: str,
        attributes_filter: str,
        exclude: bool,  # noqa: FBT001
    ) -> str:
        """Return a select pattern in string form."""
        exclude_char = "!" if exclude else ""
        return f"{exclude_char}{entities_filter}.{attributes_filter}"



"""Project Plugin Class."""

from __future__ import annotations

import copy
import typing as t

import structlog

from meltano.core.plugin.base import PluginDefinition, PluginRef, PluginType, Variant
from meltano.core.plugin.command import Command
from meltano.core.plugin.factory import base_plugin_factory
from meltano.core.plugin.requirements import PluginRequirement
from meltano.core.setting_definition import SettingDefinition
from meltano.core.utils import flatten, uniques_in

logger = structlog.stdlib.get_logger(__name__)


class CyclicInheritanceError(Exception):
    """Exception raised when project plugin inherits from itself cyclicly."""

    def __init__(self, plugin: ProjectPlugin, ancestor: ProjectPlugin):
        """Initialize cyclic inheritance error.

        Args:
            plugin: A ProjectPlugin
            ancestor: The given ProjectPlugins' ancestor.
        """
        super().__init__()

        self.plugin = plugin
        self.ancestor = ancestor

    def __str__(self) -> str:
        """Return error message.

        Returns:
            A formatted error message string.
        """
        return (
            f"{self.plugin.type.descriptor.capitalize()} '{self.plugin.name}' "
            f"cannot inherit from '{self.ancestor.name}', which itself "
            f"inherits from '{self.plugin.name}'"
        )


class ProjectPlugin(PluginRef):  # too many attrs and methods
    """ProjectPlugin class."""

    VARIANT_ATTR = "variant"

    name: str
    variant: str | None

    config_files: dict[str, str]

    def __init__(
        self,
        plugin_type: PluginType,
        name: str,
        inherit_from: str | None = None,
        namespace: str | None = None,
        variant: str | None = None,
        pip_url: str | None = None,
        python: str | None = None,
        executable: str | None = None,
        capabilities: list | None = None,
        settings_group_validation: list | None = None,
        settings: list | None = None,
        commands: dict | None = None,
        requires: dict[PluginType, list] | None = None,
        config: dict | None = None,
        default_variant=Variant.ORIGINAL_NAME,  # noqa: ANN001
        env: dict[str, str] | None = None,
        **extras,  # noqa: ANN003
    ):
        """ProjectPlugin.

        Args:
            plugin_type: PluginType instance.
            name: Plugin name.
            inherit_from: Name of plugin to inherit from.
            namespace: Plugin namespace.
            variant: Plugin variant.
            pip_url: Plugin install pip url.
            python: The python version to use for this plugin, specified as a path, or a
                executable name to find within a directory in $PATH.
            executable: Executable name.
            capabilities: Capabilities.
            settings_group_validation: Settings group validation.
            settings: Settings.
            commands: Plugin commands.
            requires: Plugin requirements.
            config: Plugin configuration.
            default_variant: Default variant for this plugin.
            env: Environment variables to inject into plugins runtime context.
            extras: Extra keyword arguments.
        """
        super().__init__(plugin_type, name)

        # Attributes will be listed in meltano.yml in the order they are set on self:
        self.inherit_from = (
            inherit_from if inherit_from and inherit_from != name else None
        )

        # If a custom definition is provided, its properties will come before
        # all others in meltano.yml
        self.custom_definition = None
        self._flattened.add("custom_definition")

        self._parent = None
        if not self.inherit_from and namespace:
            # When not explicitly inheriting, a namespace indicates an
            # embedded custom plugin definition
            self.custom_definition = PluginDefinition(
                plugin_type,
                name,
                namespace,
                variant=variant,
                pip_url=pip_url,
                python=python,
                executable=executable,
                capabilities=capabilities,
                settings_group_validation=settings_group_validation,
                settings=settings,
                requires=requires,
                **extras,
            )

            # Any properties considered "extra" by the embedded plugin definition
            # should be considered extras of the project plugin, since they are
            # the current values, not default values.
            extras = self.custom_definition.extras
            self.custom_definition.extras = {}

            # Typically, the parent is set from `ProjectPluginsService.current_plugins`,
            # where we have access to the discoverable plugin definitions coming from
            # PluginDiscoveryService, but here we can set the parent directly.
            self.parent = base_plugin_factory(self.custom_definition, variant)

        # These properties are also set on the parent, but can be overridden
        self.namespace = namespace
        self.set_presentation_attrs(extras)
        self.variant = variant
        self.pip_url = pip_url
        self.python = python
        self.executable = executable
        self.capabilities = capabilities
        self.settings_group_validation = settings_group_validation
        self.settings = list(map(SettingDefinition.parse, settings or []))
        self.commands = Command.parse_all(commands)
        self.requires = PluginRequirement.parse_all(requires)
        self.env = env or {}

        self._fallbacks.update(
            [
                "logo_url",
                "description",
                self.VARIANT_ATTR,
                "pip_url",
                "python",
                "executable",
                "capabilities",
                "settings_group_validation",
            ],
        )

        # If no variant is set, we fall back on the default
        self._defaults[self.VARIANT_ATTR] = lambda _: default_variant

        if self.inherit_from:
            # When explicitly inheriting from a project plugin or discoverable
            # definition, derive default values from our own name
            self._defaults["namespace"] = lambda plugin: plugin.name.replace("-", "_")
            self._defaults["label"] = lambda plugin: (
                f"{plugin.parent.label}: {plugin.name}"
                if plugin.parent
                else plugin.name
            )
        else:
            # When shadowing a discoverable definition with the same name (no
            # `inherit_from`), or an embedded custom definition (with
            # `namespace`), fall back on parent's values derived from its name
            # instead
            self._fallbacks.update(["namespace", "label"])

        self.config = copy.deepcopy(config or {})
        self.extras = extras

        if "profiles" in extras:
            logger.warning(
                "Plugin configuration profiles are no longer supported, ignoring "  # noqa: G004
                f"`profiles` in '{name}' {plugin_type.descriptor} definition.",
            )

    @property
    def parent(self) -> ProjectPlugin | None:
        """Plugins parent.

        Returns:
            Parent ProjectPlugin instance, or None if no parent.
        """
        return self._parent

    @parent.setter
    def parent(self, new_parent) -> None:  # noqa: ANN001
        ancestor = new_parent
        while isinstance(ancestor, self.__class__):
            if ancestor == self:
                raise CyclicInheritanceError(self, ancestor)

            ancestor = ancestor.parent

        self._parent = new_parent
        self._fallback_to = new_parent

    @property
    def is_variant_set(self) -> bool:
        """Check if variant is set explicitly.

        Returns:
            'True' if variant is set explicitly.
        """
        return self.is_attr_set(self.VARIANT_ATTR)

    @property
    def info(self) -> dict[str, str]:
        """Plugin info dict.

        Returns:
            Dictionary of plugin info (name, namespace and variant)
        """
        return {"name": self.name, "namespace": self.namespace, "variant": self.variant}

    @property
    def info_env(self) -> dict[str, str]:
        """Plugin environment info.

        Returns:
            Dictionary of plugin info formatted as Meltano environment variables.
        """
        # MELTANO_EXTRACTOR_...
        return flatten({"meltano": {self.type.singular: self.info}}, "env_var")

    @property
    def all_commands(self) -> dict[str, Command]:
        """Return all commands for this plugin.

        Returns:
            Dictionary of supported commands, including those inherited from
            the parent plugin.
        """
        return {**(self._parent.all_commands if self._parent else {}), **self.commands}

    @property
    def test_commands(self) -> dict[str, Command]:
        """Return the test commands for this plugin.

        Returns:
            Dictionary of supported test commands, including those inherited
            from the parent plugin.
        """
        return {
            name: command
            for name, command in self.all_commands.items()
            if name.startswith("test")
        }

    @property
    def supported_commands(self) -> list[str]:
        """Return supported command names.

        Returns:
            All defined command names for the plugin.
        """
        return list(self.all_commands.keys())

    def env_prefixes(self, *, for_writing=False) -> list[str]:  # noqa: ANN001
        """Return environment variable prefixes.

        Args:
            for_writing: Include parent prefix (used when writing to env vars)

        Returns:
            A list of env prefixes.
        """
        prefixes = [self.name, self.namespace]

        if for_writing:
            if self._parent:
                prefixes.extend(self._parent.env_prefixes(for_writing=True))
            prefixes.append(f"meltano_{self.type.verb}")  # MELTANO_EXTRACT_...

        return uniques_in(prefixes)

    @property
    def extra_config(self) -> dict[str, t.Any]:
        """Return plugin extra config.

        Returns:
            Dictionary of extra config.
        """
        return {f"_{key}": value for key, value in self.extras.items()}

    @property
    def config_with_extras(self) -> dict[str, t.Any]:
        """Return config with extras.

        Returns:
            Complete config dictionary, including config extras.
        """
        return {**self.config, **self.extra_config}

    @config_with_extras.setter
    def config_with_extras(self, new_config_with_extras) -> None:  # noqa: ANN001
        self.config.clear()
        self.extras.clear()

        for key, value in new_config_with_extras.items():
            if key.startswith("_"):
                self.extras[key[1:]] = value
            else:
                self.config[key] = value

    @property
    def all_settings(self) -> list[SettingDefinition]:
        """Return all settings for this plugin.

        Returns:
            List of settings, including those inherited from the parent plugin.
        """
        # New setting definitions override old ones
        new_setting_names = {setting.name for setting in self.settings}
        existing_settings = (
            [
                setting
                for setting in self._parent.all_settings
                if setting.name not in new_setting_names
            ]
            if self._parent
            else []
        )
        existing_settings.extend(self.settings)

        return [
            *existing_settings,
            *SettingDefinition.from_missing(existing_settings, self.config),
        ]

    @property
    def extra_settings(self) -> list[SettingDefinition]:
        """Return extra settings.

        Returns:
            A list of extra SettingDefinitions, including those defined by the parent.
        """
        existing_settings = self._parent.extra_settings if self._parent else []
        return [
            *existing_settings,
            *SettingDefinition.from_missing(existing_settings, self.extra_config),
        ]

    @property
    def settings_with_extras(self) -> list[SettingDefinition]:
        """Return all settings.

        Returns:
            A complete list of SettingDefinitions, including extras.
        """
        return [*self.all_settings, *self.extra_settings]

    def is_custom(self) -> bool:
        """Return if plugin is custom.

        Returns:
            'True' is plugin is custom.
        """
        return self.custom_definition is not None

    @property
    def is_shadowing(self) -> bool:
        """Return whether this plugin is shadowing a base plugin with the same name.

        Returns:
            'True' if this plugin is shadowing a base plugin with the same name.
        """
        return not self.inherit_from

    @property
    def venv_name(self) -> str:
        """Return the venv name this plugin should use.

        Returns:
            The name of this plugins parent if both pip urls are the same, else
            this plugins name.
        """
        if not self.inherit_from:
            return self.name

        if not self.pip_url or (self.parent.pip_url == self.pip_url):
            return self.parent.name

        return self.name

    def get_requirements(
        self,
        plugin_types: t.Iterable[PluginType] | None = None,
    ) -> dict[PluginType, list[PluginRequirement]]:
        """Return the requirements for this plugin.

        Args:
            plugin_types: The plugin types to include.

        Returns:
            A list of requirements for this plugin, optionally filtered for
            specified plugin types.
        """
        plugin_types = plugin_types or list(PluginType)
        return {
            plugin_type: [
                *(
                    self._parent.all_requires.get(plugin_type, [])
                    if self._parent
                    else []
                ),
                *self.requires.get(plugin_type, []),
            ]
            for plugin_type in plugin_types
        }

    @property
    def all_requires(self) -> dict[PluginType, list]:
        """Return all requires for this plugin.

        Returns:
            List of supported requires, including those inherited from the
            parent plugin.
        """
        return self.get_requirements(plugin_types=None)

    @property
    def requirements(self) -> list[ProjectPlugin]:
        """Return the requirements for this plugin.

        Returns:
            A list of requirements for this plugin.
        """
        return [
            ProjectPlugin(plugin_type=plugin_type, name=dep.name, variant=dep.variant)
            for plugin_type, deps in self.all_requires.items()
            for dep in deps
        ]



"""Environment context for the Snowplow tracker."""

from __future__ import annotations

import os
import platform
import typing as t
import uuid
from collections import defaultdict
from contextlib import suppress
from datetime import datetime, timezone
from functools import cached_property
from pathlib import Path
from warnings import warn

import psutil
from snowplow_tracker import SelfDescribingJson
from structlog.stdlib import get_logger

import meltano
from meltano.core.tracking.schemas import EnvironmentContextSchema
from meltano.core.utils import get_boolean_env_var, hash_sha256, safe_hasattr, strtobool

logger = get_logger(__name__)

# This file is only ever created in CI when building a release
release_marker_path = Path(__file__).parent / ".release_marker"


def _get_parent_context_uuid_str() -> str | None:
    with suppress(KeyError):
        uuid_str = os.environ["MELTANO_PARENT_CONTEXT_UUID"]
        try:
            return str(uuid.UUID(uuid_str))
        except ValueError:
            warn(
                (
                    f"Invalid telemetry parent environment context UUID "
                    f"{uuid_str!r} from $MELTANO_PARENT_CONTEXT_UUID - "
                    "Meltano will continue as if $MELTANO_PARENT_CONTEXT_UUID "
                    "had not been set"
                ),
                stacklevel=2,
            )
    return None


class EnvironmentContext(SelfDescribingJson):
    """Environment context for the Snowplow tracker."""

    ci_markers: t.ClassVar[set[str]] = {
        "GITHUB_ACTIONS",
        "CI",
    }
    notable_flag_env_vars: t.ClassVar[set[str]] = {
        "CODESPACES",
        *ci_markers,
    }
    notable_hashed_env_vars: t.ClassVar[set[str]] = {
        "CODESPACE_NAME",
        "GITHUB_REPOSITORY",
        "GITHUB_USER",
    }

    @classmethod
    def _notable_hashed_env_vars(cls) -> t.Iterable[str]:
        for env_var_name in cls.notable_hashed_env_vars:
            with suppress(KeyError):  # Skip unset env vars
                env_var_value = os.environ[env_var_name]
                yield env_var_name, hash_sha256(env_var_value)

    @classmethod
    def _notable_flag_env_vars(cls) -> t.Iterable[str]:
        for env_var_name in cls.notable_flag_env_vars:
            with suppress(KeyError):  # Skip unset env vars
                env_var_value = os.environ[env_var_name]
                try:
                    yield env_var_name, strtobool(env_var_value)
                except ValueError:
                    yield env_var_name, None

    def __init__(self) -> None:
        """Initialize the environment context."""
        super().__init__(
            EnvironmentContextSchema.url,
            {
                "context_uuid": str(uuid.uuid4()),
                "parent_context_uuid": _get_parent_context_uuid_str(),
                "meltano_version": meltano.__version__,
                "is_dev_build": not release_marker_path.exists(),
                "is_ci_environment": any(
                    get_boolean_env_var(marker) for marker in self.ci_markers
                ),
                "notable_flag_env_vars": dict(self._notable_flag_env_vars()),
                "notable_hashed_env_vars": dict(self._notable_hashed_env_vars()),
                "python_version": platform.python_version(),
                "python_implementation": platform.python_implementation(),
                **self.system_info,
                **self.process_info,
            },
        )

    @cached_property
    def system_info(self) -> dict[str, t.Any]:
        """Get system information.

        Returns:
            A dictionary containing system information.
        """
        try:
            freedesktop_data = platform.freedesktop_os_release()
        except Exception:
            freedesktop_data = defaultdict(type(None))

        return {
            "system_name": platform.system() or None,
            "system_release": platform.release() or None,
            "system_version": platform.version() or None,
            "machine": platform.machine() or None,
            "windows_edition": platform.win32_edition()
            if safe_hasattr(platform, "win32_edition")
            else None,
            "freedesktop_id": freedesktop_data["ID"],
            "freedesktop_id_like": freedesktop_data.get("ID_LIKE", None),
            "freedesktop_version_id": freedesktop_data.get("VERSION_ID", None),
        }

    @staticmethod
    def get_process_timestamp(process: psutil.Process) -> str:
        """Obtain the creation time of a process as a ISO 8601 timestamp.

        Args:
            process: The process to obtain the creation time from.

        Returns:
            A ISO 8601 timestamp formatted string.
        """
        return f"{datetime.fromtimestamp(process.create_time(), tz=timezone.utc).isoformat()}"  # noqa: E501

    @cached_property
    def process_info(self) -> dict[str, t.Any]:
        """Obtain the process information for the current process.

        Returns:
            A dictionary containing the process information. Such as the hashed
            process name, pid, core counts, etc
        """
        process = psutil.Process()
        with process.oneshot():
            return {
                "num_cpu_cores": psutil.cpu_count(),
                "num_cpu_cores_available": self.num_available_cores,
                "process_hierarchy": [
                    {
                        "process_name_hash": hash_sha256(proc.name()),
                        "process_creation_timestamp": self.get_process_timestamp(proc),
                    }
                    for proc in (process, *process.parents())
                ],
            }

    @cached_property
    def num_available_cores(self) -> int:
        """Obtain the number of available CPU cores.

        Uses `sched_getaffinity` where available, otherwise falls back to
        `cpu_count`.

        Returns:
            The number of available CPU cores.
        """
        if safe_hasattr(os, "sched_getaffinity"):
            return len(os.sched_getaffinity(0))
        return os.cpu_count()


environment_context = EnvironmentContext()



from __future__ import annotations

import platform
import shutil
import stat
from pathlib import Path

import pytest

from meltano.core.project_init_service import (
    ProjectInitService,
    ProjectInitServiceError,
)


@pytest.mark.parametrize(
    "create_project_dir",
    (
        False,
        True,
    ),
    ids=(
        "new directory",
        "empty directory",
    ),
)
def test_project_init_success(
    *,
    create_project_dir: bool,
    tmp_path: Path,
    pushd,
) -> None:
    projects_dir = tmp_path.joinpath("success")
    projects_dir.mkdir()
    pushd(projects_dir)

    project_dir = projects_dir.joinpath("test_project")

    if create_project_dir:
        project_dir.mkdir()

    ProjectInitService(project_dir.relative_to(Path.cwd())).init(activate=False)

    shutil.rmtree(project_dir)

    if create_project_dir:
        project_dir.mkdir()

    ProjectInitService(project_dir.absolute()).init(activate=False)


def test_project_init_non_empty_directory(tmp_path: Path, pushd) -> None:
    projects_dir = tmp_path.joinpath("exists")
    projects_dir.mkdir()
    pushd(projects_dir)

    project_dir = projects_dir.joinpath("test_project")
    project_dir.joinpath("test").mkdir(parents=True)
    ProjectInitService(project_dir).init(activate=False)


def test_project_init_existing_meltano_yml(tmp_path: Path, pushd) -> None:
    projects_dir = tmp_path.joinpath("exists")
    projects_dir.mkdir()
    pushd(projects_dir)

    project_dir = projects_dir.joinpath("test_project")
    project_dir.mkdir(parents=True)
    project_dir.joinpath("meltano.yml").touch()

    with pytest.raises(
        ProjectInitServiceError,
        match=(
            "A `meltano.yml` file already exists in the target directory. "
            "Use `--force` to overwrite it."
        ),
    ):
        ProjectInitService(project_dir).init(activate=False)

    ProjectInitService(project_dir).init(activate=False, force=True)


def test_project_init_no_write_permission(tmp_path: Path, pushd) -> None:
    if platform.system() == "Windows":
        pytest.xfail(
            "Windows can still create new directories inside a read-only directory.",
        )

    protected_dir = tmp_path.joinpath("protected")
    protected_dir.mkdir()
    # read and execute, but not write
    protected_dir.chmod(stat.S_IREAD | stat.S_IEXEC)
    pushd(protected_dir)

    project_dir = protected_dir.joinpath("test_project")

    with pytest.raises(
        ProjectInitServiceError,
        match="Permission denied to create 'test_project'",
    ):
        ProjectInitService(project_dir).init(activate=False)


def test_project_init_missing_parent_directory(tmp_path: Path, pushd) -> None:
    if platform.system() == "Windows":
        pytest.xfail(
            "Windows can't remove a directory that is in use. "
            "See https://docs.python.org/3/library/os.html#os.remove",
        )

    missing_dir = tmp_path.joinpath("missing")
    missing_dir.mkdir()
    pushd(missing_dir)

    project_dir = missing_dir.joinpath("test_project")

    project_init_service = ProjectInitService(project_dir)
    missing_dir.rmdir()  # remove the parent directory
    with pytest.raises(
        ProjectInitServiceError,
        match="Could not create directory 'test_project'.",
    ):
        project_init_service.init(activate=False)



"""Project plugin service."""

from __future__ import annotations

import enum
import typing as t
from contextlib import contextmanager, suppress
from functools import cached_property

import structlog

from meltano.core.error import MeltanoError
from meltano.core.locked_definition_service import LockedDefinitionService
from meltano.core.plugin import PluginRef, PluginType
from meltano.core.plugin.error import PluginNotFoundError, PluginParentNotFoundError
from meltano.core.plugin_lock_service import PluginLockService

if t.TYPE_CHECKING:
    from meltano.core.environment import EnvironmentPluginConfig
    from meltano.core.plugin.project_plugin import ProjectPlugin
    from meltano.core.project import Project

logger = structlog.stdlib.get_logger(__name__)


class DefinitionSource(enum.Flag):
    """The source of a plugin definition."""

    NONE = 0
    HUB = enum.auto()
    CUSTOM = enum.auto()
    LOCKFILE = enum.auto()
    INHERITED = enum.auto()

    ANY = HUB | CUSTOM | LOCKFILE | INHERITED
    LOCAL = ~HUB


class PluginAlreadyAddedException(Exception):
    """Raised when a plugin is already added to the project."""

    def __init__(self, plugin: PluginRef, new_plugin: PluginRef):
        """Create a new Plugin Already Added Exception.

        Args:
            plugin: The plugin that was already added.
            new_plugin: The plugin that was attempted to be added.
        """
        self.plugin = plugin
        self.new_plugin = new_plugin
        super().__init__()


class PluginDefinitionNotFoundError(MeltanoError):
    """Raised when no plugin definition is found."""

    def __init__(
        self,
        plugin: ProjectPlugin,
        error: Exception | None,
        source: DefinitionSource,
    ):
        """Initialize a new error.

        Args:
            plugin: The plugin that was not found.
            error: The error that was raised, if any.
            source: The sources searched for the plugin.
        """
        reason = (
            str(error)
            if error
            else f"No definition found for {plugin.type.descriptor} {plugin.name}"
        )
        instruction = None

        if DefinitionSource.HUB in source:
            instruction = "Check https://hub.meltano.com/ for available plugins"
        else:
            instruction = (
                "Try running `meltano lock --update --all` to ensure your plugins are "
                "up to date"
            )

        super().__init__(reason=reason, instruction=instruction)


class AmbiguousMappingName(MeltanoError):
    """Occurs when the same mapping name is used in multiple mappers."""

    def __init__(self, mapping_name: str):
        """Initialize the exception.

        Args:
            mapping_name: The name of the schedule that does not exist.
        """
        super().__init__(
            reason=f"Ambiguous mapping name {mapping_name}, found multiple matches.",
            instruction=(
                "Alter one of the instances of this mapping name to make it distinct."
            ),
        )


class ProjectPluginsService:  # (too many methods, attributes)
    """Project Plugins Service."""

    def __init__(self, project: Project):
        """Create a new Project Plugins Service.

        Args:
            project: The Meltano project.
        """
        self.project = project
        self.lock_service = PluginLockService(project)
        self.locked_definition_service = LockedDefinitionService(project)
        self._prefer_source = DefinitionSource.LOCAL

    @cached_property
    def current_plugins(self):  # noqa: ANN201
        """Return the current plugins.

        Returns:
            The current plugins.
        """
        return self.project.config_service.current_meltano_yml.plugins

    @contextmanager
    def update_plugins(self):  # noqa: ANN201
        """Update the current plugins.

        Yields:
            The current plugins.
        """
        with self.project.config_service.update_meltano_yml() as meltano_yml:
            yield meltano_yml.plugins

    def add_to_file(self, plugin: ProjectPlugin):  # noqa: ANN201
        """Add plugin to `meltano.yml`.

        Args:
            plugin: The plugin to add.

        Raises:
            PluginAlreadyAddedException: If the plugin is already added.

        Returns:
            The added plugin.
        """
        # FIXME: `should_add_to_file` is a method from `BasePlugin`, which is
        #        not a subclass of `ProjectPlugin`. I've left this call to it
        #        in-place because I'm worried that removing it will break stuff
        #        that relies on it, but something is definitely wrong here.
        #        We default to `True` for `ProjectPlugin` objects.
        if not getattr(plugin, "should_add_to_file", lambda: True)():
            return plugin

        with suppress(PluginNotFoundError):
            existing_plugin = self.get_plugin(plugin)
            raise PluginAlreadyAddedException(existing_plugin, plugin)

        with self.update_plugins() as plugins:
            if plugin.type not in plugins:
                plugins[plugin.type] = []

            plugins[plugin.type].append(plugin)

        return plugin

    def remove_from_file(self, plugin: ProjectPlugin):  # noqa: ANN201
        """Remove plugin from `meltano.yml`.

        Args:
            plugin: The plugin to remove.

        Returns:
            The removed plugin.
        """
        # Will raise if the plugin isn't actually in the file
        self.get_plugin(plugin)

        with self.update_plugins() as plugins:
            plugins[plugin.type].remove(plugin)

        return plugin

    def has_plugin(self, plugin_name: str) -> bool:
        """Check if plugin exists for the given name.

        Args:
            plugin_name: The name of the plugin to check for.

        Returns:
            True if the plugin exists, False otherwise.
        """
        try:
            self.find_plugin(plugin_name)
            return True
        except PluginNotFoundError:
            return False

    def find_plugin(
        self,
        plugin_name: str,
        plugin_type: PluginType | None = None,
        invokable=None,  # noqa: ANN001
        configurable=None,  # noqa: ANN001
    ) -> ProjectPlugin:
        """Find a plugin.

        Args:
            plugin_name: The name of the plugin to find.
            plugin_type: Optionally the type of plugin.
            invokable: Optionally limit the search to invokable plugins.
            configurable: Optionally limit the search to configurable plugins.

        Returns:
            The plugin.

        Raises:
            PluginNotFoundError: If the plugin is not found.
        """
        if "@" in plugin_name:
            plugin_name, profile_name = plugin_name.split("@", 2)
            logger.warning(
                "Plugin configuration profiles are no longer supported, "  # noqa: G004
                f"ignoring `@{profile_name}` in plugin name.",
            )

        for plugin in self.plugins(ensure_parent=False):
            if (
                plugin.name == plugin_name  # (with too much logic)
                and (plugin_type is None or plugin.type == plugin_type)
                and (
                    invokable is None
                    or self.ensure_parent(plugin).is_invokable() == invokable
                )
                and (
                    configurable is None
                    or self.ensure_parent(plugin).is_configurable() == configurable
                )
            ):
                return self.ensure_parent(plugin)

            if plugin.type == PluginType.MAPPERS:
                mapping = self._find_mapping(plugin_name, plugin)
                if mapping:
                    return mapping
        raise PluginNotFoundError(
            PluginRef(plugin_type, plugin_name) if plugin_type else plugin_name,
        )

    def _find_mapping(self, plugin_name: str, plugin: ProjectPlugin) -> ProjectPlugin:
        mapping_name = plugin.extra_config.get("_mapping_name")
        if mapping_name == plugin_name:  # noqa: RET503
            all_mappings = self.find_plugins_by_mapping_name(mapping_name)
            if len(all_mappings) > 1:
                raise AmbiguousMappingName(mapping_name)
            return self.ensure_parent(plugin)

    def find_plugin_by_namespace(
        self,
        plugin_type: PluginType,
        namespace: str,
    ) -> ProjectPlugin:
        """Find a plugin based on its PluginType and namespace.

        For example, PluginType.EXTRACTORS and namespace tap_custom
        will return the extractor for the tap-custom plugin.

        Args:
            plugin_type: The type of plugin to find.
            namespace: The namespace of the plugin.

        Returns:
            The plugin if found.

        Raises:
            PluginNotFoundError: If no plugin is found.
        """
        try:
            return next(
                plugin
                for plugin in self.plugins()
                if plugin.namespace == namespace and plugin_type == plugin.type
            )
        except StopIteration as stop:
            raise PluginNotFoundError(namespace) from stop

    def find_plugins_by_mapping_name(self, mapping_name: str) -> list[ProjectPlugin]:
        """Find plugins with the specified mapping name in their mappings config.

        Args:
            mapping_name: The name of the mapping to find.

        Returns:
            The mapping plugins with the specified mapping name.

        Raises:
            PluginNotFoundError: If no mapper plugin with the specified mapping
                name is found.
        """
        if found := [
            plugin
            for plugin in self.get_plugins_of_type(plugin_type=PluginType.MAPPERS)
            if plugin.extra_config.get("_mapping_name") == mapping_name
        ]:
            return found
        raise PluginNotFoundError(mapping_name)

    def get_plugin(self, plugin_ref: PluginRef) -> ProjectPlugin:
        """Get a plugin using its PluginRef.

        Args:
            plugin_ref: The plugin reference to use.

        Returns:
            The plugin if found.

        Raises:
            PluginNotFoundError: If the plugin is not found.
        """
        try:
            plugin = next(
                plugin
                for plugin in self.plugins(ensure_parent=False)
                if plugin == plugin_ref
            )

            return self.ensure_parent(plugin)
        except StopIteration as stop:
            raise PluginNotFoundError(plugin_ref) from stop

    def get_plugins_of_type(
        self,
        plugin_type: PluginType,
        *,
        ensure_parent=True,  # noqa: ANN001
    ) -> list[ProjectPlugin]:
        """Return plugins of specified type.

        Args:
            plugin_type: The type of the plugins to return.
            ensure_parent: If True, ensure that plugin has a parent plugin set.

        Returns:
            A list of plugins of the specified plugin type.
        """
        plugins = self.current_plugins[plugin_type]

        if ensure_parent:
            for plugin in plugins:
                self.ensure_parent(plugin)

        return plugins

    def plugins_by_type(self, *, ensure_parent=True):  # noqa: ANN001, ANN201
        """Return plugins grouped by type.

        Args:
            ensure_parent: If True, ensure that plugin has a parent plugin set.

        Returns:
            A dict of plugins grouped by type.
        """
        return {
            plugin_type: self.get_plugins_of_type(
                plugin_type,
                ensure_parent=ensure_parent,
            )
            for plugin_type in PluginType
        }

    def plugins(self, *, ensure_parent=True) -> t.Generator[ProjectPlugin, None, None]:  # noqa: ANN001
        """Return all plugins.

        Args:
            ensure_parent: If True, ensure that plugin has a parent plugin set.

        Yields:
            Plugins.
        """
        yield from (
            plugin
            for _, plugins in self.plugins_by_type(ensure_parent=ensure_parent).items()
            for plugin in plugins
        )

    def update_plugin(self, plugin: ProjectPlugin, *, keep_config: bool = False):  # noqa: ANN201
        """Update a plugin.

        Args:
            plugin: The plugin to update.
            keep_config: Whether to keep the previous configuration for the updated
                plugin.

        Returns:
            A tuple containing the updated and outdated plugins.

        Raises:
            PluginNotFoundError: If the plugin is not found.
        """
        with self.update_plugins() as plugins:
            # find the proper plugin to update
            try:
                idx, outdated = next(
                    (idx, plg)
                    for idx, plg in enumerate(plugins[plugin.type])
                    if plg == plugin
                )

                if keep_config:
                    plugin.config_with_extras = outdated.config_with_extras

                plugins[plugin.type][idx] = plugin

                return plugin, outdated
            except StopIteration as stop:
                raise PluginNotFoundError(plugin) from stop

    def update_environment_plugin(self, plugin: EnvironmentPluginConfig) -> None:
        """Update a plugin configuration inside a Meltano environment.

        Args:
            plugin: The plugin configuration to update.
        """
        with self.project.config_service.update_active_environment() as environment:
            environment.config.plugins.setdefault(plugin.type, [])

            # find the proper plugin to update
            p_idx, p_outdated = next(
                (
                    (idx, plg)
                    for idx, plg in enumerate(environment.config.plugins[plugin.type])
                    if plg == plugin
                ),
                (None, None),
            )

            if p_idx is None:
                environment.config.plugins.setdefault(plugin.type, [])
                environment.config.plugins[plugin.type].append(plugin)
            else:
                environment.config.plugins[plugin.type][p_idx] = plugin

    def _get_parent_from_hub(self, plugin: ProjectPlugin) -> ProjectPlugin:
        """Get the parent plugin from the hub.

        Args:
            plugin: The plugin to get the parent of.

        Returns:
            The parent plugin.

        Raises:
            PluginParentNotFoundError: If the parent plugin is not found.
        """
        try:
            return self.project.hub_service.get_base_plugin(
                plugin,
                variant_name=plugin.variant,
            )
        except PluginNotFoundError as err:
            if plugin.inherit_from:
                raise PluginParentNotFoundError(plugin, err) from err
            raise

    def find_parent(
        self,
        plugin: ProjectPlugin,
    ) -> tuple[ProjectPlugin, DefinitionSource]:
        """Find the parent plugin of a plugin.

        Args:
            plugin: The plugin to find the parent of.

        Returns:
            The parent plugin and the source of the parent.

        Raises:
            PluginDefinitionNotFoundError: If the parent plugin is not found.
        """
        error = None
        if (
            plugin.inherit_from
            and not plugin.is_variant_set
            and DefinitionSource.INHERITED in self._prefer_source
        ):
            try:
                return (
                    self.find_plugin(
                        plugin_type=plugin.type,
                        plugin_name=plugin.inherit_from,
                    ),
                    DefinitionSource.INHERITED,
                )
            except PluginNotFoundError as inherited_exc:
                error = inherited_exc

        if DefinitionSource.LOCKFILE in self._prefer_source:
            try:
                return (
                    self.locked_definition_service.get_base_plugin(
                        plugin,
                        variant_name=plugin.variant,
                    ),
                    DefinitionSource.LOCKFILE,
                )
            except PluginNotFoundError as lockfile_exc:
                error = lockfile_exc

        if DefinitionSource.HUB in self._prefer_source:
            try:
                return (self._get_parent_from_hub(plugin), DefinitionSource.HUB)
            except Exception as hub_exc:
                error = hub_exc

        raise PluginDefinitionNotFoundError(
            plugin,
            error,
            self._prefer_source,
        ) from error

    def get_parent(self, plugin: ProjectPlugin) -> ProjectPlugin:
        """Get plugin's parent plugin.

        Args:
            plugin: The plugin to get the parent of.

        Returns:
            The parent plugin or None if the plugin has no parent.
        """
        parent, source = self.find_parent(plugin)

        logger.debug(
            "Found plugin parent",
            plugin=plugin.name,
            parent=parent.name,
            source=source.name,
        )
        return parent

    def ensure_parent(self, plugin: ProjectPlugin) -> ProjectPlugin:
        """Ensure that plugin has a parent set.

        Args:
            plugin: To set the parent of if necessary.

        Returns:
            The plugin (updated if necessary).
        """
        if not plugin.parent:
            plugin.parent = self.get_parent(plugin)

        return plugin

    def get_transformer(self) -> ProjectPlugin:
        """Get first available Transformer plugin.

        Raises:
            PluginNotFoundError: If there is no transformer.

        Returns:
            First available transformer plugin.
        """
        if transformer := next(
            iter(self.get_plugins_of_type(plugin_type=PluginType.TRANSFORMERS)),
            None,
        ):
            return transformer
        raise PluginNotFoundError("No Plugin of type Transformer found.")  # noqa: EM101

    @contextmanager
    def use_preferred_source(self, source: DefinitionSource) -> None:
        """Prefer a source of definition.

        Args:
            source: The source to prefer.

        Yields:
            None.
        """
        previous = self._prefer_source
        self._prefer_source = source
        yield
        self._prefer_source = previous
